package at.bestsolution.rsd.rt.client.opentracing;

import java.util.ArrayDeque;
import java.util.Deque;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.TreeSet;

import org.eclipse.microprofile.config.ConfigProvider;
import org.osgi.service.component.annotations.Component;
import org.osgi.service.component.annotations.Reference;
import org.osgi.service.component.annotations.ReferenceCardinality;
import org.osgi.service.component.annotations.ReferencePolicy;
import org.osgi.service.component.annotations.ReferencePolicyOption;

import at.bestsolution.rsd.rt.interceptor.RemoteCallInterceptor;
import at.bestsolution.rsd.rt.interceptor.RemoteResultInterceptor;
import at.bestsolution.rsd.rt.opentracing.TraceProvider;
import io.opentracing.Scope;
import io.opentracing.Span;
import io.opentracing.SpanContext;
import io.opentracing.Tracer;
import io.opentracing.log.Fields;
import io.opentracing.propagation.Format;
import io.opentracing.propagation.TextMap;
import io.opentracing.tag.Tags;
import io.opentracing.util.GlobalTracer;

@Component
public class OpenTracingClientInterceptor implements RemoteCallInterceptor, RemoteResultInterceptor {
	private static class ActiveScope {
		Scope scope;
		Span span;
		
		public ActiveScope(Scope scope, Span span) {
			this.scope = scope;
			this.span = span;
		}
	}
	
	private static final ThreadLocal<Deque<ActiveScope>> ACTIVE_SCOPE = ThreadLocal.withInitial( () -> new ArrayDeque<>());
	
	private String getTraceProviderId() {
		return ConfigProvider.getConfig().getOptionalValue("rsd.opentracing.provider.id", String.class)
				.orElse("jaeger");
	}
	
	private static class OrderedItem implements Comparable<OrderedItem> {
		private final int rank;
		private final TraceProvider provider;
		
		public OrderedItem(TraceProvider provider, int rank) {
			this.provider = provider;
			this.rank = rank;
		}
		
		@Override
		public int compareTo(OrderedItem o) {
			return Integer.compare(rank, o.rank);
		}
	}
	
	private final TreeSet<OrderedItem> reg = new TreeSet<>();
	
	@Reference(cardinality = ReferenceCardinality.MULTIPLE, policy = ReferencePolicy.DYNAMIC, policyOption = ReferencePolicyOption.GREEDY)
	public void registerTraceProvider(TraceProvider tracer, Map<String, Object> properties) {
		synchronized (reg) {
			reg.add(new OrderedItem(tracer, (Integer)properties.getOrDefault("service.ranking", 0)));	
		}
		
	}
	
	public void unregisterTraceProvider(TraceProvider tracer) {
		reg.removeIf( item -> item.provider == tracer);
	}
	
	public TraceProvider getActiveTracer() {
		synchronized (reg) {
			if( this.reg.isEmpty() ) {
				return new TraceProvider() {
					
					@Override
					public Tracer getTracer() {
						return GlobalTracer.get();
					}
					
					@Override
					public String getId() {
						return "<unknown>";
					}
				};
			} else {
				String providerId = getTraceProviderId();
				
				if( providerId != null ) {
					Optional<TraceProvider> provider = this.reg
						.stream()
						.map( item -> item.provider)
						.filter( p -> providerId.equals(p.getId()))
						.findFirst();
					
					if( provider.isPresent() ) {
						return provider.get();
					}
				}
				
				return this.reg.first().provider;
			}			
		}
	}
	
	@Override
	public String name() {
		return "Opentracing-Interceptor";
	}
	
	// https://github.com/eclipse/microprofile-opentracing/blob/master/spec/src/main/asciidoc/microprofile-opentracing.asciidoc
	@Override
	public boolean preInvokation(RemoteCallInterceptor.CallContext context) {
		try {
			TraceProvider provider = getActiveTracer();
			Tracer tracer = provider.getTracer();
			
			Tracer.SpanBuilder spanBuilder = tracer.buildSpan( context.getMethod().split("#")[1] );
			spanBuilder = spanBuilder.withTag(Tags.COMPONENT.getKey(), context.getComponentType());
			spanBuilder = spanBuilder.withTag(context.getComponentType() + ".bean", context.getMethod().split("#")[0]);
			spanBuilder = spanBuilder.withTag(Tags.SPAN_KIND.getKey(), Tags.SPAN_KIND_CLIENT);
			
			Span activeSpan = tracer.activeSpan();
			if( activeSpan != null ) {
				spanBuilder.asChildOf(activeSpan);
			}
			
			Span span = spanBuilder.start();
			Scope scope = tracer.activateSpan(span);
			SpanContext ctx = span.context();
				
			Map<String, String> map = new HashMap<>();
				
			tracer.inject(ctx, Format.Builtin.TEXT_MAP, new TextMap() {
				
				@Override
				public void put(String key, String value) {
					map.put(key, value);
				}
				
				@Override
				public Iterator<Entry<String, String>> iterator() {
					throw new IllegalStateException();
				}
			});
			
			context.setProperty("__rsd_trace-span-context", map);
			context.setProperty("__rsd_trace-provider-id", provider.getId());
			
			ACTIVE_SCOPE.get().push(new ActiveScope(scope, span));
		} catch( Throwable t ) {
			t.printStackTrace();
		}
		
		return true;
	}
	

	@Override
	public Object process(ResultContext context, Object result, Throwable t) {
		ActiveScope activeScope = ACTIVE_SCOPE.get().poll();
		if( activeScope != null ) {
			activeScope.scope.close();
			
			Span span = activeScope.span;
			if( t != null ) {
				Tags.ERROR.set(span, true);
				Map<String, Object> errorEvent = new HashMap<>();
				errorEvent.put(Fields.EVENT, "error");
				errorEvent.put(Fields.ERROR_OBJECT, t);
				errorEvent.put(Fields.MESSAGE, t.getMessage());
				span.log(errorEvent);
			}
			span.finish();
		}
		
		return result;
	}
}
